from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, ClassVar, Literal, TypeAlias, cast

from typing_extensions import Self

from cognite.client.data_classes._base import (
    CogniteObject,
    CognitePrimitiveUpdate,
    CogniteResource,
    CogniteResourceList,
    CogniteUpdate,
    ExternalIDTransformerMixin,
    PropertySpec,
    UnknownCogniteObject,
    WriteableCogniteResource,
    WriteableCogniteResourceList,
)

if TYPE_CHECKING:
    from cognite.client import CogniteClient

TargetStatus: TypeAlias = Literal["running", "paused"]
JobStatus: TypeAlias = Literal[
    "startup_error",
    "paused",
    "shutting_down",
    "connected",
    "running",
    "waiting",
    "connection_error",
    "cdf_write_error",
    "transform_error",
]


@dataclass
class JobFormat(CogniteObject, ABC):
    _type: ClassVar[str]

    @classmethod
    @abstractmethod
    def _load_job(cls, resource: dict[str, Any]) -> Self:
        raise NotImplementedError()

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        type_ = resource.get("type")
        if type_ is None and hasattr(cls, "_type"):
            type_ = cls._type
        elif type_ is None:
            raise KeyError("type")
        job_cls = _JOBFORMAT_CLASS_BY_TYPE.get(type_)
        if job_cls is None:
            return UnknownCogniteObject(resource)  # type: ignore[return-value]
        return cast(Self, job_cls._load_job(resource))

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case)
        output["type"] = self._type
        return output


@dataclass
class Prefix(CogniteObject):
    from_topic: bool | None = None
    prefix: str | None = None


@dataclass
class ValueFormat(JobFormat):
    _type = "value"
    encoding: Literal["utf16", "utf16le"] | None = None
    compression: str | None = None
    prefix: Prefix | None = None

    @classmethod
    def _load_job(cls, resource: dict[str, Any]) -> ValueFormat:
        return cls(
            encoding=resource.get("encoding"),
            compression=resource.get("compression"),
            prefix=Prefix._load(resource["prefix"]) if "prefix" in resource else None,
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case)
        if self.prefix:
            output["prefix"] = self.prefix.dump(camel_case)
        return output


@dataclass
class RockwellFormat(JobFormat):
    _type = "rockwell"
    encoding: Literal["utf16", "utf16le"] | None = None
    compression: str | None = None
    prefix: Prefix | None = None

    @classmethod
    def _load_job(cls, resource: dict[str, Any]) -> RockwellFormat:
        return cls(
            encoding=resource.get("encoding"),
            compression=resource.get("compression"),
            prefix=Prefix._load(resource["prefix"]) if "prefix" in resource else None,
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case)
        if self.prefix:
            output["prefix"] = self.prefix.dump(camel_case)
        return output


@dataclass
class CustomFormat(JobFormat):
    _type = "custom"
    mapping_id: str
    encoding: Literal["utf16", "utf16le"] | None = None
    compression: str | None = None

    @classmethod
    def _load_job(cls, resource: dict[str, Any]) -> CustomFormat:
        return cls(
            encoding=resource.get("encoding", None),
            compression=resource.get("compression"),
            mapping_id=resource["mappingId"],
        )


@dataclass
class CogniteFormat(JobFormat):
    _type = "cognite"
    encoding: Literal["utf16", "utf16le"] | None = None
    compression: str | None = None
    prefix: Prefix | None = None

    @classmethod
    def _load_job(cls, resource: dict[str, Any]) -> CogniteFormat:
        return cls(
            encoding=resource.get("encoding", None),
            compression=resource.get("compression"),
            prefix=Prefix._load(resource["prefix"]) if "prefix" in resource else None,
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case)
        if self.prefix:
            output["prefix"] = self.prefix.dump(camel_case)
        return output


@dataclass
class JobConfig(CogniteObject, ABC):
    @classmethod
    @abstractmethod
    def _load_config(cls, data: dict[str, Any]) -> Self:
        raise NotImplementedError()

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        if "topicFilter" in resource:
            return cast(Self, MQTTConfig._load_config(resource))
        elif "topic" in resource:
            return cast(Self, KafkaConfig._load_config(resource))
        elif "interval" in resource and "path" in resource:
            return cast(Self, RestConfig._load_config(resource))
        else:
            return cast(Self, UnknownCogniteObject(resource))


@dataclass
class MQTTConfig(JobConfig):
    topic_filter: str

    @classmethod
    def _load_config(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(topic_filter=resource["topicFilter"])


@dataclass
class KafkaConfig(JobConfig):
    topic: str
    partitions: int = 1

    @classmethod
    def _load_config(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(topic=resource["topic"], partitions=resource.get("partitions", 1))


@dataclass
class IncrementalLoad(CogniteObject, ABC):
    _type: ClassVar[str]

    @classmethod
    @abstractmethod
    def _load_incremental_load(cls, resource: dict[str, Any]) -> Self:
        raise NotImplementedError()

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        type_ = resource.get("type")
        if type_ is None and hasattr(cls, "_type"):
            type_ = cls._type
        elif type_ is None:
            raise KeyError("type")
        incremental_load_cls = _INCREMENTALLOAD_CLASS_BY_TYPE.get(type_)
        if incremental_load_cls is None:
            return UnknownCogniteObject(resource)  # type: ignore[return-value]
        return cast(Self, incremental_load_cls._load_incremental_load(resource))

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case)
        output["type"] = self._type
        return output


@dataclass
class BodyLoad(IncrementalLoad):
    _type = "body"
    value: str

    @classmethod
    def _load_incremental_load(cls, resource: dict[str, Any]) -> Self:
        return cls(value=resource["value"])


@dataclass
class HeaderValueLoad(IncrementalLoad):
    _type = "headerValue"
    key: str
    value: str

    @classmethod
    def _load_incremental_load(cls, resource: dict[str, Any]) -> Self:
        return cls(key=resource["key"], value=resource["value"])


@dataclass
class QueryParamLoad(IncrementalLoad):
    _type = "queryParam"
    key: str
    value: str

    @classmethod
    def _load_incremental_load(cls, resource: dict[str, Any]) -> Self:
        return cls(key=resource["key"], value=resource["value"])


@dataclass
class NextUrlLoad(IncrementalLoad):
    _type = "nextUrl"
    value: str

    @classmethod
    def _load_incremental_load(cls, resource: dict[str, Any]) -> Self:
        return cls(value=resource["value"])


@dataclass
class RestConfig(JobConfig):
    interval: Literal["5m", "15m", "1h", "6h", "12h", "1d"]
    path: str
    method: Literal["get", "post"] = "get"
    body: str | None = None
    query: dict[str, str] | None = None
    headers: dict[str, str] | None = None
    incremental_load: IncrementalLoad | None = None
    pagination: IncrementalLoad | None = None

    def __post_init__(self) -> None:
        if isinstance(self.incremental_load, NextUrlLoad):
            raise ValueError("incremental_load cannot be of type NextUrlLoad")

    @classmethod
    def _load_config(cls, resource: dict[str, Any]) -> Self:
        return cls(
            interval=resource["interval"],
            path=resource["path"],
            method=resource.get("method", "get"),
            body=resource.get("body"),
            query=resource.get("query"),
            headers=resource.get("headers"),
            incremental_load=IncrementalLoad._load(resource["incrementalLoad"])
            if "incrementalLoad" in resource
            else None,
            pagination=IncrementalLoad._load(resource["pagination"]) if "pagination" in resource else None,
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case)
        if self.incremental_load:
            output["incrementalLoad"] = self.incremental_load.dump(camel_case)
        if self.pagination:
            output["pagination"] = self.pagination.dump(camel_case)
        return output


class _JobCore(WriteableCogniteResource["JobWrite"]):
    def __init__(
        self, external_id: str, destination_id: str, source_id: str, format: JobFormat, config: JobConfig | None = None
    ) -> None:
        self.external_id = external_id
        self.destination_id = destination_id
        self.source_id = source_id
        self.format = format
        self.config = config

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case)
        output["format"] = self.format.dump(camel_case)
        if self.config is not None:
            output["config"] = self.config.dump(camel_case)
        return output


class JobWrite(_JobCore):
    """A hosted extractor job represents the running extractor.

    Jobs produce logs and metrics that give the state of the job.

    This is the write/request format of the job.

    Args:
        external_id (str): The external ID provided by the client. Must be unique for the resource type.
        destination_id (str): ID of the destination this job should write to.
        source_id (str): ID of the source this job should read from.
        format: The format of the messages from the source. This is used to convert
            messages coming from the source system to a format that can be inserted into CDF.
        config: Configuration for the job. This must match the source. For example, if the source is MQTT,
            this must be an MQTT configuration.

    """

    def as_write(self) -> JobWrite:
        return self

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> JobWrite:
        return cls(
            external_id=resource["externalId"],
            destination_id=resource["destinationId"],
            source_id=resource["sourceId"],
            format=JobFormat._load(resource["format"]),
            config=JobConfig._load(resource["config"]) if "config" in resource else None,
        )


class Job(_JobCore):
    """A hosted extractor job represents the running extractor.

    Jobs produce logs and metrics that give the state of the job.

    This is the read/response format of the job.

    Args:
        external_id (str): The external ID provided by the client. Must be unique for the resource type.
        destination_id (str): ID of the destination this job should write to.
        source_id (str): ID of the source this job should read from.
        format (JobFormat): The format of the messages from the source. This is used to convert messages coming from the source system to a format that can be inserted into CDF.
        target_status (TargetStatus): The target status of a job. Set this to start or stop the job.
        status (JobStatus): Status of this job.
        created_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        last_updated_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        config (JobConfig | None): Configuration for the job. This is specific to the source system.

    """

    def __init__(
        self,
        external_id: str,
        destination_id: str,
        source_id: str,
        format: JobFormat,
        target_status: TargetStatus,
        status: JobStatus,
        created_time: int,
        last_updated_time: int,
        config: JobConfig | None = None,
    ) -> None:
        super().__init__(external_id, destination_id, source_id, format, config)
        self.target_status = target_status
        self.status = status
        self.created_time = created_time
        self.last_updated_time = last_updated_time

    def as_write(self) -> JobWrite:
        return JobWrite(
            external_id=self.external_id,
            destination_id=self.destination_id,
            source_id=self.source_id,
            format=self.format,
            config=self.config,
        )

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Job:
        return cls(
            external_id=resource["externalId"],
            destination_id=resource["destinationId"],
            source_id=resource["sourceId"],
            format=JobFormat._load(resource["format"]),
            target_status=resource["targetStatus"],
            status=resource["status"],
            config=JobConfig._load(resource["config"]) if "config" in resource else None,
            created_time=resource["createdTime"],
            last_updated_time=resource["lastUpdatedTime"],
        )


class JobWriteList(CogniteResourceList[JobWrite], ExternalIDTransformerMixin):
    _RESOURCE = JobWrite


class JobList(WriteableCogniteResourceList[JobWrite, Job], ExternalIDTransformerMixin):
    _RESOURCE = Job

    def as_write(self) -> JobWriteList:
        return JobWriteList([job.as_write() for job in self.data])


class JobUpdate(CogniteUpdate):
    def __init__(self, external_id: str) -> None:
        super().__init__(external_id=external_id)

    class _StringUpdate(CognitePrimitiveUpdate):
        def set(self, value: str) -> JobUpdate:
            return self._set(value)

    class _FormatUpdate(CognitePrimitiveUpdate):
        def set(self, value: JobFormat) -> JobUpdate:
            return self._set(value.dump(camel_case=True))

    class _ConfigUpdate(CognitePrimitiveUpdate):
        def set(self, value: JobConfig) -> JobUpdate:
            return self._set(value.dump(camel_case=True))

    class _TargetStatusUpdate(CognitePrimitiveUpdate):
        def set(self, value: TargetStatus) -> JobUpdate:
            return self._set(value)

    @property
    def destination_id(self) -> JobUpdate._StringUpdate:
        return self._StringUpdate(self, "destinationId")

    @property
    def source_id(self) -> JobUpdate._StringUpdate:
        return self._StringUpdate(self, "sourceId")

    @property
    def format(self) -> JobUpdate._FormatUpdate:
        return self._FormatUpdate(self, "format")

    @property
    def target_status(self) -> JobUpdate._TargetStatusUpdate:
        return self._TargetStatusUpdate(self, "targetStatus")

    @property
    def config(self) -> JobUpdate._ConfigUpdate:
        return self._ConfigUpdate(self, "config")

    @classmethod
    def _get_update_properties(cls, item: CogniteResource | None = None) -> list[PropertySpec]:
        return [
            PropertySpec("destination_id", is_nullable=False),
            PropertySpec("source_id", is_nullable=False),
            PropertySpec("format", is_nullable=False),
            PropertySpec("target_status", is_nullable=False),
            PropertySpec("config", is_nullable=False),
        ]


class JobLogs(CogniteResource):
    """Logs for a hosted extractor job.

    Args:
        job_external_id (str): The external ID of the job.
        type (Literal['paused', 'startup_error', 'connection_error', 'connected', 'transform_error', 'cdf_write_error', 'ok']): Type of log entry.
        created_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        message (str | None): Log message. Not all log entries have messages.

    Statuses

    * `paused`: indicates that the job has been stopped manually.
    * `startup_error`: indicates that the job failed to start at all and requires changes to configuration.
    * `connection_error`: indicates that the job lost connection to the source system.
    * `connected`: indicates that the job has connected to the source but did not yet receive any data,
    * `transform_error`: indicates that the job received data, but it failed to transform.
    * `cdf_write_error`: indicates that ingesting the data to CDF failed.
    * `ok`: means that data was successfully ingested to CDF.

    """

    def __init__(
        self,
        job_external_id: str,
        type: Literal[
            "paused", "startup_error", "connection_error", "connected", "transform_error", "cdf_write_error", "ok"
        ],
        created_time: int,
        message: str | None = None,
    ) -> None:
        self.job_external_id = job_external_id
        self.type = type
        self.created_time = created_time
        self.message = message

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> JobLogs:
        return cls(
            job_external_id=resource["jobExternalId"],
            type=resource["type"],
            created_time=resource["createdTime"],
            message=resource.get("message"),
        )


class JobMetrics(CogniteResource):
    """Metrics for a hosted extractor job.

    Args:
        job_external_id (str): External ID of the job this metrics batch belongs to.
        timestamp (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds. Metrics are from the UTC hour this timestamp is ingest. For example, if this timestamp is at 01:43:15, the metrics batch contains metrics from 01:00:00 to 01:43:15.
        source_messages (int): Number of messages received from the source system.
        cdf_input_values (int): Destination resources successfully transformed and passed to CDF.
        cdf_requests (int): Requests made to CDF containing data produced by this job.
        transform_failures (int): Source messages that failed to transform.
        cdf_write_failures (int): Times the destination received data from transformations, but failed to produce a valid request to CDF.
        cdf_skipped_values (int): Values the destination received from the source, then decided to skip due to data type mismatch, invalid content, or other.
        cdf_failed_values (int): Values the destination was unable to upload to CDF.
        cdf_uploaded_values (int): Values the destination successfully uploaded to CDF.

    """

    def __init__(
        self,
        job_external_id: str,
        timestamp: int,
        source_messages: int,
        cdf_input_values: int,
        cdf_requests: int,
        transform_failures: int,
        cdf_write_failures: int,
        cdf_skipped_values: int,
        cdf_failed_values: int,
        cdf_uploaded_values: int,
    ) -> None:
        self.job_external_id = job_external_id
        self.timestamp = timestamp
        self.source_messages = source_messages
        self.cdf_input_values = cdf_input_values
        self.cdf_requests = cdf_requests
        self.transform_failures = transform_failures
        self.cdf_write_failures = cdf_write_failures
        self.cdf_skipped_values = cdf_skipped_values
        self.cdf_failed_values = cdf_failed_values
        self.cdf_uploaded_values = cdf_uploaded_values

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            job_external_id=resource["jobExternalId"],
            timestamp=resource["timestamp"],
            source_messages=resource["sourceMessages"],
            cdf_input_values=resource["cdfInputValues"],
            cdf_requests=resource["cdfRequests"],
            transform_failures=resource["transformFailures"],
            cdf_write_failures=resource["cdfWriteFailures"],
            cdf_skipped_values=resource["cdfSkippedValues"],
            cdf_failed_values=resource["cdfFailedValues"],
            cdf_uploaded_values=resource["cdfUploadedValues"],
        )


class JobLogsList(CogniteResourceList[JobLogs]):
    _RESOURCE = JobLogs


class JobMetricsList(CogniteResourceList[JobMetrics]):
    _RESOURCE = JobMetrics


_JOBFORMAT_CLASS_BY_TYPE: dict[str, type[JobFormat]] = {
    subclass._type: subclass  # type: ignore[type-abstract]
    for subclass in JobFormat.__subclasses__()
}

_INCREMENTALLOAD_CLASS_BY_TYPE: dict[str, type[IncrementalLoad]] = {
    subclass._type: subclass  # type: ignore[type-abstract]
    for subclass in IncrementalLoad.__subclasses__()
}



from __future__ import annotations

from cognite.client.data_classes.postgres_gateway.tables import (
    Column,
    ColumnList,
    ColumnType,
    RawTable,
    RawTableOptions,
    RawTableWrite,
    Table,
    TableList,
    TableWrite,
    TableWriteList,
    ViewTable,
    ViewTableWrite,
)
from cognite.client.data_classes.postgres_gateway.users import (
    SessionCredentials,
    User,
    UserCreated,
    UserCreatedList,
    UserList,
    UserUpdate,
    UserWrite,
    UserWriteList,
)

__all__ = [
    "Column",
    "ColumnList",
    "ColumnType",
    "RawTable",
    "RawTableOptions",
    "RawTableWrite",
    "SessionCredentials",
    "Table",
    "TableList",
    "TableWrite",
    "TableWriteList",
    "User",
    "UserCreated",
    "UserCreatedList",
    "UserList",
    "UserUpdate",
    "UserWrite",
    "UserWriteList",
    "ViewTable",
    "ViewTableWrite",
]



from __future__ import annotations

from collections.abc import Mapping, Sequence
from copy import deepcopy
from typing import TYPE_CHECKING, Any, Literal, cast, overload

from cognite.client._api_client import APIClient
from cognite.client._constants import DEFAULT_LIMIT_READ
from cognite.client.data_classes import Annotation, AnnotationFilter, AnnotationList, AnnotationUpdate
from cognite.client.data_classes._base import CogniteResource, PropertySpec, T_CogniteResource
from cognite.client.data_classes.annotations import AnnotationCore, AnnotationReverseLookupFilter, AnnotationWrite
from cognite.client.data_classes.contextualization import ResourceReference, ResourceReferenceList
from cognite.client.utils._auxiliary import is_unlimited, split_into_chunks
from cognite.client.utils._experimental import FeaturePreviewWarning
from cognite.client.utils._identifier import IdentifierSequence
from cognite.client.utils._text import convert_all_keys_to_camel_case
from cognite.client.utils._validation import assert_type

if TYPE_CHECKING:
    from cognite.client import CogniteClient
    from cognite.client.config import ClientConfig


class AnnotationsAPI(APIClient):
    _RESOURCE_PATH = "/annotations"

    def __init__(self, config: ClientConfig, api_version: str | None, cognite_client: CogniteClient) -> None:
        super().__init__(config, api_version, cognite_client)
        self._reverse_lookup_warning = FeaturePreviewWarning(
            api_maturity="beta", sdk_maturity="beta", feature_name="Annotation reverse lookup"
        )

    @overload
    def create(self, annotations: Annotation | AnnotationWrite) -> Annotation: ...

    @overload
    def create(self, annotations: Sequence[Annotation | AnnotationWrite]) -> AnnotationList: ...

    def create(
        self, annotations: Annotation | AnnotationWrite | Sequence[Annotation | AnnotationWrite]
    ) -> Annotation | AnnotationList:
        """`Create annotations <https://developer.cognite.com/api#tag/Annotations/operation/annotationsCreate>`_

        Args:
            annotations (Annotation | AnnotationWrite | Sequence[Annotation | AnnotationWrite]): Annotation(s) to create

        Returns:
            Annotation | AnnotationList: Created annotation(s)
        """
        assert_type(annotations, "annotations", [AnnotationCore, Sequence])

        return self._create_multiple(
            list_cls=AnnotationList,
            resource_cls=Annotation,
            resource_path=self._RESOURCE_PATH + "/",
            items=annotations,
            input_resource_cls=AnnotationWrite,
        )

    @overload
    def suggest(self, annotations: Annotation) -> Annotation: ...

    @overload
    def suggest(self, annotations: Sequence[Annotation]) -> AnnotationList: ...

    def suggest(self, annotations: Annotation | Sequence[Annotation]) -> Annotation | AnnotationList:
        """`Suggest annotations <https://developer.cognite.com/api#tag/Annotations/operation/annotationsSuggest>`_

        Args:
            annotations (Annotation | Sequence[Annotation]): annotation(s) to suggest. They must have status set to "suggested".

        Returns:
            Annotation | AnnotationList: suggested annotation(s)
        """
        assert_type(annotations, "annotations", [Annotation, Sequence])
        # Deal with status fields in both cases: Single item and list of items
        items: list[dict[str, Any]] | dict[str, Any] = (
            [self._sanitize_suggest_item(ann) for ann in annotations]
            if isinstance(annotations, Sequence)
            else self._sanitize_suggest_item(annotations)
        )
        return self._create_multiple(
            list_cls=AnnotationList,
            resource_cls=Annotation,
            resource_path=self._RESOURCE_PATH + "/suggest",
            items=items,
        )

    @staticmethod
    def _sanitize_suggest_item(annotation: Annotation | dict[str, Any]) -> dict[str, Any]:
        # Check that status is set to suggested if it is set and afterwards remove it
        item = annotation.dump(camel_case=True) if isinstance(annotation, Annotation) else deepcopy(annotation)
        if "status" in item:
            if item["status"] != "suggested":
                raise ValueError("status field for Annotation suggestions must be set to 'suggested'")
            del item["status"]
        return item

    @classmethod
    def _convert_resource_to_patch_object(
        cls,
        resource: CogniteResource,
        update_attributes: list[PropertySpec],
        mode: Literal["replace_ignore_null", "patch", "replace"] = "replace_ignore_null",
        cdf_item_by_id: Mapping[Any, T_CogniteResource] | None = None,
    ) -> dict[str, dict[str, dict]]:
        if not isinstance(resource, Annotation):
            return APIClient._convert_resource_to_patch_object(resource, update_attributes)
        annotation: Annotation = resource

        assert annotation.id is not None
        annotation_update = AnnotationUpdate(id=annotation.id)
        for attr in update_attributes:
            getattr(annotation_update, attr.name).set(getattr(annotation, attr.name))
        return annotation_update.dump()

    @overload
    def update(
        self,
        item: Annotation | AnnotationWrite | AnnotationUpdate,
        mode: Literal["replace_ignore_null", "patch", "replace"] = "replace_ignore_null",
    ) -> Annotation: ...

    @overload
    def update(
        self,
        item: Sequence[Annotation | AnnotationWrite | AnnotationUpdate],
        mode: Literal["replace_ignore_null", "patch", "replace"] = "replace_ignore_null",
    ) -> AnnotationList: ...

    def update(
        self,
        item: Annotation
        | AnnotationWrite
        | AnnotationUpdate
        | Sequence[Annotation | AnnotationWrite | AnnotationUpdate],
        mode: Literal["replace_ignore_null", "patch", "replace"] = "replace_ignore_null",
    ) -> Annotation | AnnotationList:
        """`Update annotations <https://developer.cognite.com/api#tag/Annotations/operation/annotationsUpdate>`_

        Args:
            item (Annotation | AnnotationWrite | AnnotationUpdate | Sequence[Annotation | AnnotationWrite | AnnotationUpdate]): Annotation or list of annotations to update (or patch or list of patches to apply)
            mode (Literal['replace_ignore_null', 'patch', 'replace']): How to update data when a non-update object is given (Annotation or -Write). If you use 'replace_ignore_null', only the fields you have set will be used to replace existing (default). Using 'replace' will additionally clear all the fields that are not specified by you. Last option, 'patch', will update only the fields you have set and for container-like fields such as metadata or labels, add the values to the existing. For more details, see :ref:`appendix-update`.

        Returns:
            Annotation | AnnotationList: No description."""
        return self._update_multiple(
            list_cls=AnnotationList, resource_cls=Annotation, update_cls=AnnotationUpdate, items=item, mode=mode
        )

    def delete(self, id: int | Sequence[int]) -> None:
        """`Delete annotations <https://developer.cognite.com/api#tag/Annotations/operation/annotationsDelete>`_

        Args:
            id (int | Sequence[int]): ID or list of IDs to be deleted
        """
        self._delete_multiple(identifiers=IdentifierSequence.load(ids=id), wrap_ids=True)

    def retrieve_multiple(self, ids: Sequence[int]) -> AnnotationList:
        """`Retrieve annotations by IDs <https://developer.cognite.com/api#tag/Annotations/operation/annotationsByids>`_`

        Args:
            ids (Sequence[int]): list of IDs to be retrieved

        Returns:
            AnnotationList: list of annotations
        """
        identifiers = IdentifierSequence.load(ids=ids, external_ids=None)
        return self._retrieve_multiple(list_cls=AnnotationList, resource_cls=Annotation, identifiers=identifiers)

    def retrieve(self, id: int) -> Annotation | None:
        """`Retrieve an annotation by id <https://developer.cognite.com/api#tag/Annotations/operation/annotationsGet>`_

        Args:
            id (int): id of the annotation to be retrieved

        Returns:
            Annotation | None: annotation requested
        """
        identifiers = IdentifierSequence.load(ids=id, external_ids=None).as_singleton()
        return self._retrieve_multiple(list_cls=AnnotationList, resource_cls=Annotation, identifiers=identifiers)

    def reverse_lookup(self, filter: AnnotationReverseLookupFilter, limit: int | None = None) -> ResourceReferenceList:
        """Reverse lookup annotated resources based on having annotations matching the filter.

        Args:
            filter (AnnotationReverseLookupFilter): Filter to apply
            limit (int | None): Maximum number of results to return. Defaults to None (all).

        Returns:
            ResourceReferenceList: List of resource references

        Examples:

            Retrieve the first 100 ids of annotated resources mathing the 'file' resource type:

                >>> from cognite.client import CogniteClient
                >>> from cognite.client.data_classes import AnnotationReverseLookupFilter
                >>> client = CogniteClient()
                >>> flt = AnnotationReverseLookupFilter(annotated_resource_type="file")
                >>> res = client.annotations.reverse_lookup(flt, limit=100)
        """
        self._reverse_lookup_warning.warn()
        assert_type(filter, "filter", types=[AnnotationReverseLookupFilter], allow_none=False)

        return self._list(
            list_cls=ResourceReferenceList,
            resource_cls=ResourceReference,
            method="POST",
            limit=limit,
            filter=filter.dump(camel_case=True),
            url_path=self._RESOURCE_PATH + "/reverselookup",
            api_subversion="beta",
        )

    def list(self, filter: AnnotationFilter | dict, limit: int | None = DEFAULT_LIMIT_READ) -> AnnotationList:
        """`List annotations. <https://developer.cognite.com/api#tag/Annotations/operation/annotationsFilter>`_

        Note:
            Passing a filter with both 'annotated_resource_type' and 'annotated_resource_ids' is always required.

        Args:
            filter (AnnotationFilter | dict): Return annotations with parameter values that match what is specified.
            limit (int | None): Maximum number of annotations to return. Defaults to 25. Set to -1, float("inf") or None to return all items.

        Returns:
            AnnotationList: list of annotations

        Example:

            List all annotations for the file with id=123:

                >>> from cognite.client import CogniteClient
                >>> from cognite.client.data_classes import AnnotationFilter
                >>> client = CogniteClient()
                >>> flt = AnnotationFilter(annotated_resource_type="file", annotated_resource_ids=[{"id": 123}])
                >>> res = client.annotations.list(flt, limit=None)
        """
        assert_type(filter, "filter", [AnnotationFilter, dict], allow_none=False)

        if isinstance(filter, AnnotationFilter):
            filter = filter.dump(camel_case=True)
        elif isinstance(filter, dict):
            filter = convert_all_keys_to_camel_case(filter)

        if "annotatedResourceIds" not in filter or "annotatedResourceType" not in filter:
            raise ValueError("Both 'annotated_resource_type' and 'annotated_resource_ids' are required in filter!")
        res_ids = list(map(convert_all_keys_to_camel_case, filter.pop("annotatedResourceIds")))

        remaining_limit = limit
        is_finite_limit = not is_unlimited(limit)

        all_annots = AnnotationList([], cognite_client=self._cognite_client)
        for id_chunk in split_into_chunks(res_ids, 1000):
            filter["annotatedResourceIds"] = id_chunk
            chunk_result = self._list(
                list_cls=AnnotationList, resource_cls=Annotation, method="POST", limit=remaining_limit, filter=filter
            )
            all_annots.extend(chunk_result)

            if is_finite_limit:
                remaining_limit = cast(int, limit) - len(all_annots)
                if remaining_limit == 0:
                    break
        return all_annots



from __future__ import annotations

from cognite.client._cognite_client import CogniteClient
from cognite.client._constants import _RUNNING_IN_BROWSER
from cognite.client._version import __version__
from cognite.client.config import ClientConfig, global_config
from cognite.client.data_classes import data_modeling

__all__ = ["ClientConfig", "CogniteClient", "__version__", "data_modeling", "global_config"]

if _RUNNING_IN_BROWSER:
    from cognite.client.utils._pyodide_helpers import patch_sdk_for_pyodide

    patch_sdk_for_pyodide()



from __future__ import annotations

import time
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Literal, cast

from typing_extensions import Self

from cognite.client.data_classes._base import (
    CogniteObject,
    CogniteResource,
    CogniteResourceList,
    ExternalIDTransformerMixin,
    IdTransformerMixin,
    WriteableCogniteResource,
    WriteableCogniteResourceList,
)
from cognite.client.data_classes.simulators.logs import SimulatorLog
from cognite.client.utils._experimental import FeaturePreviewWarning
from cognite.client.utils._importing import local_import
from cognite.client.utils._retry import Backoff
from cognite.client.utils._text import to_snake_case

if TYPE_CHECKING:
    import pandas

    from cognite.client import CogniteClient

_WARNING = FeaturePreviewWarning(api_maturity="General Availability", sdk_maturity="alpha", feature_name="Simulators")


@dataclass
class SimulationValueUnitName(CogniteObject):
    name: str | None = None

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            name=resource.get("name"),
        )

    def __post_init__(self) -> None:
        _WARNING.warn()


@dataclass
class SimulationValueUnit(SimulationValueUnitName):
    external_id: str | None = None

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            external_id=resource.get("externalId"),
            name=resource.get("name"),
        )

    def __post_init__(self) -> None:
        _WARNING.warn()


@dataclass
class SimulationInputOverride(CogniteObject):
    reference_id: str
    value: str | int | float | list[str] | list[int] | list[float]
    unit: SimulationValueUnitName | None = None

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            reference_id=resource["referenceId"],
            value=resource["value"],
            unit=(SimulationValueUnitName._load(resource["unit"], cognite_client) if resource.get("unit") else None),
        )

    def __post_init__(self) -> None:
        _WARNING.warn()

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case=camel_case)
        if self.unit is not None:
            output["unit"] = self.unit.dump(camel_case=camel_case)

        return output


class SimulationRunCore(WriteableCogniteResource["SimulationRunWrite"]):
    def __init__(
        self,
        run_type: str | None,
        routine_external_id: str,
        run_time: int | None = None,
    ) -> None:
        self.run_type = run_type
        self.run_time = run_time
        self.routine_external_id = routine_external_id


class SimulationRunWrite(SimulationRunCore):
    """
    Request to run a simulator routine asynchronously.

    Args:
        routine_external_id (str): External id of the associated simulator routine
        run_type (str | None): The type of the simulation run
        run_time (int | None): Run time in milliseconds. Reference timestamp used for data pre-processing and data sampling.
        queue (bool | None): Queue the simulation run when connector is down.
        log_severity (str | None): Override the minimum severity level for the simulation run logs. If not provided, the minimum severity is read from the connector logger configuration.
        inputs (list[SimulationInputOverride] | None): List of input overrides
    """

    def __init__(
        self,
        routine_external_id: str,
        run_type: str | None = None,
        run_time: int | None = None,
        queue: bool | None = None,
        log_severity: str | None = None,
        inputs: list[SimulationInputOverride] | None = None,
    ) -> None:
        super().__init__(
            routine_external_id=routine_external_id,
            run_type=run_type,
            run_time=run_time,
        )
        self.queue = queue
        self.log_severity = log_severity
        self.inputs = inputs

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> SimulationRunWrite:
        inputs = resource.get("inputs", None)
        return cls(
            run_type=resource.get("runType"),
            routine_external_id=resource["routineExternalId"],
            run_time=resource.get("runTime"),
            queue=resource.get("queue"),
            log_severity=resource.get("logSeverity"),
            inputs=([SimulationInputOverride._load(_input) for _input in inputs] if inputs else None),
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case=camel_case)
        if self.inputs is not None:
            output["inputs"] = [input_.dump(camel_case=camel_case) for input_ in self.inputs]
        return output

    def as_write(self) -> SimulationRunWrite:
        return self


class SimulationRun(SimulationRunCore):
    """
    Every time a simulation routine executes, a simulation run object is created.

    This object ensures that each execution of a routine is documented and traceable.
    Each run has an associated simulation data resource, which stores the inputs and outputs of a
    simulation run, capturing the values set into and read from the simulator model to ensure
    the traceability and integrity of the simulation data.

    Simulation runs provide a historical record of the simulations performed, allowing users to analyze
    and compare different runs, track changes over time, and make informed decisions based on the simulation results.

    This is the read/response format of a simulation run.

    Args:
        id (int): The id of the simulation run
        simulator_external_id (str): External id of the associated simulator
        simulator_integration_external_id (str): External id of the associated simulator integration
        model_external_id (str): External id of the associated simulator model
        model_revision_external_id (str): External id of the associated simulator model revision
        routine_revision_external_id (str): External id of the associated simulator routine revision
        routine_external_id (str): External id of the associated simulator routine
        run_type (Literal['external', 'manual', 'scheduled']): The type of the simulation run
        status (Literal['ready', 'running', 'success', 'failure']): The status of the simulation run
        data_set_id (int): The id of the dataset associated with the simulation run
        user_id (str): The id of the user who executed the simulation run
        log_id (int): The id of the log associated with the simulation run
        created_time (int): The number of milliseconds since epoch
        last_updated_time (int): The number of milliseconds since epoch
        status_message (str | None): The status message of the simulation run
        simulation_time (int | None): Simulation time in milliseconds. Timestamp when the input data was sampled. Used for indexing input and output time series.
        run_time (int | None): Run time in milliseconds. Reference timestamp used for data pre-processing and data sampling.
        cognite_client (CogniteClient | None): An optional CogniteClient to associate with this data class.

    """

    def __init__(
        self,
        id: int,
        simulator_external_id: str,
        simulator_integration_external_id: str,
        model_external_id: str,
        model_revision_external_id: str,
        routine_revision_external_id: str,
        routine_external_id: str,
        run_type: Literal["external", "manual", "scheduled"],
        status: Literal["ready", "running", "success", "failure"],
        data_set_id: int,
        user_id: str,
        log_id: int,
        created_time: int,
        last_updated_time: int,
        status_message: str | None = None,
        simulation_time: int | None = None,
        run_time: int | None = None,
        cognite_client: CogniteClient | None = None,
    ) -> None:
        super().__init__(
            run_type=run_type,
            routine_external_id=routine_external_id,
            run_time=run_time,
        )
        self.id = id
        self.created_time = created_time
        self.last_updated_time = last_updated_time
        self.simulator_external_id = simulator_external_id
        self.simulator_integration_external_id = simulator_integration_external_id
        self.model_external_id = model_external_id
        self.model_revision_external_id = model_revision_external_id
        self.routine_revision_external_id = routine_revision_external_id
        self.simulation_time = simulation_time
        self.status = status
        self.status_message = status_message
        self.data_set_id = data_set_id
        self.user_id = user_id
        self.log_id = log_id
        self._cognite_client = cast("CogniteClient", cognite_client)

    def as_write(self) -> SimulationRunWrite:
        return SimulationRunWrite(
            routine_external_id=self.routine_external_id,
            run_type=self.run_type,
            run_time=self.run_time,
        )

    def get_logs(self) -> SimulatorLog | None:
        """`Retrieve logs for this simulation run. <https://developer.cognite.com/api#tag/Simulator-Logs/operation/simulator_logs_by_ids_simulators_logs_byids_post>`_

        Returns:
            SimulatorLog | None: Log for the simulation run.
        """
        return self._cognite_client.simulators.logs.retrieve(ids=self.log_id)

    def get_data(self) -> SimulationRunDataItem | None:
        """`Retrieve data associated with this simulation run. <https://developer.cognite.com/api#tag/Simulation-Runs/operation/simulation_data_by_run_id_simulators_runs_data_list_post>`_

        Returns:
            SimulationRunDataItem | None: Data for the simulation run.
        """
        data = self._cognite_client.simulators.runs.list_run_data(run_id=self.id)
        if data:
            return data[0]

        return None

    def update(self) -> None:
        """Update the simulation run object to the latest state. Useful if the run was created with wait=False."""
        # same logic as Cognite Functions
        latest = self._cognite_client.simulators.runs.retrieve(ids=self.id)
        if latest is None:
            raise RuntimeError("Unable to update the simulation run object (it was not found)")
        else:
            self.status = latest.status
            self.status_message = latest.status_message
            self.simulation_time = latest.simulation_time
            self.last_updated_time = latest.last_updated_time

    def wait(self, timeout: float = 60) -> None:
        """Waits for simulation status to become either success or failure.
        This is generally not needed to call directly, as client.simulators.routines.run(...) will wait for the simulation to finish by default.

        Args:
            timeout (float): Time out after this many seconds. Defaults to 60 seconds.
        """

        end_time = time.time() + timeout
        poll_backoff = Backoff(max_wait=30, base=10)

        while time.time() < end_time:
            self.update()
            if self.status in ["success", "failure"]:
                return
            sleep_time = min(next(poll_backoff) + 1, end_time - time.time())
            time.sleep(sleep_time)

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> SimulationRun:
        return cls(
            id=resource["id"],
            created_time=resource["createdTime"],
            last_updated_time=resource["lastUpdatedTime"],
            simulator_external_id=resource["simulatorExternalId"],
            simulator_integration_external_id=resource["simulatorIntegrationExternalId"],
            model_external_id=resource["modelExternalId"],
            model_revision_external_id=resource["modelRevisionExternalId"],
            routine_external_id=resource["routineExternalId"],
            routine_revision_external_id=resource["routineRevisionExternalId"],
            status=resource["status"],
            data_set_id=resource["dataSetId"],
            run_type=resource["runType"],
            user_id=resource["userId"],
            log_id=resource["logId"],
            status_message=resource.get("statusMessage"),
            run_time=resource.get("runTime"),
            simulation_time=resource.get("simulationTime"),
            cognite_client=cognite_client,
        )


class SimulationValueBase(CogniteObject):
    """
    Base class for simulation values. This class is used to define the value and its type.
    The value can be a string, double, array of strings or array of doubles.
    """

    def __init__(
        self,
        reference_id: str,
        value: str | int | float | list[str] | list[int] | list[float],
        value_type: Literal["STRING", "DOUBLE", "STRING_ARRAY", "DOUBLE_ARRAY"],
        unit: SimulationValueUnitName | None = None,
        simulator_object_reference: dict[str, str] | None = None,
        timeseries_external_id: str | None = None,
    ) -> None:
        self.reference_id = reference_id
        self.value = value
        self.value_type = value_type
        self.unit = unit
        self.simulator_object_reference = simulator_object_reference
        self.timeseries_external_id = timeseries_external_id

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            reference_id=resource["referenceId"],
            value=resource["value"],
            value_type=resource["valueType"],
            unit=SimulationValueUnitName._load(resource["unit"], cognite_client) if resource.get("unit") else None,
            simulator_object_reference=resource.get("simulatorObjectReference"),
            timeseries_external_id=resource.get("timeseriesExternalId"),
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case=camel_case)
        if self.unit is not None:
            output["unit"] = self.unit.dump(camel_case=camel_case)
        if self.simulator_object_reference is not None:
            output["simulatorObjectReference"] = self.simulator_object_reference
        if self.timeseries_external_id is not None:
            output["timeseriesExternalId"] = self.timeseries_external_id
        return output


class SimulationInput(SimulationValueBase):
    """
    This class is used to define the value and its type.
    The value can be a string, double, array of strings or array of doubles.
    """

    def __init__(
        self,
        reference_id: str,
        value: str | int | float | list[str] | list[int] | list[float],
        value_type: Literal["STRING", "DOUBLE", "STRING_ARRAY", "DOUBLE_ARRAY"],
        unit: SimulationValueUnitName | None = None,
        simulator_object_reference: dict[str, str] | None = None,
        timeseries_external_id: str | None = None,
        overridden: bool | None = None,
    ) -> None:
        super().__init__(
            reference_id=reference_id,
            value=value,
            value_type=value_type,
            unit=unit,
            simulator_object_reference=simulator_object_reference,
            timeseries_external_id=timeseries_external_id,
        )
        self.overridden = overridden

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            reference_id=resource["referenceId"],
            value=resource["value"],
            value_type=resource["valueType"],
            unit=SimulationValueUnitName._load(resource["unit"], cognite_client) if resource.get("unit") else None,
            simulator_object_reference=resource.get("simulatorObjectReference"),
            timeseries_external_id=resource.get("timeseriesExternalId"),
            overridden=resource.get("overridden"),
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case=camel_case)
        if self.unit is not None:
            output["unit"] = self.unit.dump(camel_case=camel_case)
        if self.simulator_object_reference is not None:
            output["simulatorObjectReference"] = self.simulator_object_reference
        if self.timeseries_external_id is not None:
            output["timeseriesExternalId"] = self.timeseries_external_id
        if self.overridden is not None:
            output["overridden"] = self.overridden
        return output


class SimulationOutput(SimulationValueBase):
    """
    This class is used to return the outputs generated during the simulation.
    The value can be a string, double, array of strings or array of doubles.
    """

    pass


class SimulationRunDataItem(CogniteResource):
    def __init__(
        self,
        run_id: int,
        inputs: list[SimulationInput],
        outputs: list[SimulationOutput],
    ) -> None:
        super().__init__()
        self.run_id = run_id
        self.inputs = inputs
        self.outputs = outputs

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        inputs = [SimulationInput._load(input_) for input_ in resource["inputs"]]
        outputs = [SimulationOutput._load(output_) for output_ in resource["outputs"]]
        return cls(
            run_id=resource["runId"],
            inputs=inputs,
            outputs=outputs,
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case=camel_case)
        output["inputs"] = [input_.dump(camel_case=camel_case) for input_ in self.inputs]
        output["outputs"] = [output_.dump(camel_case=camel_case) for output_ in self.outputs]
        return output

    def to_pandas(  # type: ignore [override]
        self,
    ) -> pandas.DataFrame:
        """Convert the simulation run data to a pandas DataFrame.

        Returns:
            pandas.DataFrame: The dataframe.
        """
        pd = local_import("pandas")

        def _create_row(item: SimulationInput | SimulationOutput) -> dict:
            """Create a row dictionary for an input or output item."""
            item_type = (isinstance(item, SimulationInput) and "Input") or "Output"
            row = {
                "run_id": self.run_id,
                "type": item_type,
                "reference_id": item.reference_id,
                "value": item.value,
                "unit_name": item.unit.name if item.unit else None,
                "value_type": item.value_type,
                "overridden": getattr(item, "overridden", None),
                "timeseries_external_id": item.timeseries_external_id,
            }

            if item.simulator_object_reference:
                for reference_key, reference_value in item.simulator_object_reference.items():
                    snake_key = to_snake_case(reference_key)
                    row[snake_key] = reference_value

            return row

        rows = [_create_row(item) for item in self.inputs + self.outputs]

        return pd.DataFrame(rows)


class SimulatorRunDataList(CogniteResourceList[SimulationRunDataItem], IdTransformerMixin):
    _RESOURCE = SimulationRunDataItem

    def to_pandas(  # type: ignore [override]
        self,
    ) -> pandas.DataFrame:
        """Convert the simulation run data list to a pandas DataFrame.

        Returns:
            pandas.DataFrame: The dataframe.
        """
        pd = local_import("pandas")
        return pd.concat([item.to_pandas() for item in self.data], ignore_index=True)


class SimulationRunWriteList(CogniteResourceList[SimulationRunWrite], ExternalIDTransformerMixin):
    _RESOURCE = SimulationRunWrite


class SimulatorRunList(WriteableCogniteResourceList[SimulationRunWrite, SimulationRun], IdTransformerMixin):
    _RESOURCE = SimulationRun

    def as_write(self) -> SimulationRunWriteList:
        return SimulationRunWriteList([a.as_write() for a in self.data], cognite_client=self._get_cognite_client())



from __future__ import annotations

from collections.abc import Iterator, Sequence
from typing import TYPE_CHECKING, Literal, overload

from cognite.client._api_client import APIClient
from cognite.client._constants import DEFAULT_LIMIT_READ
from cognite.client.data_classes import (
    BoundingBox3D,
    ThreeDAssetMapping,
    ThreeDAssetMappingList,
    ThreeDAssetMappingWrite,
    ThreeDModel,
    ThreeDModelList,
    ThreeDModelRevision,
    ThreeDModelRevisionList,
    ThreeDModelRevisionUpdate,
    ThreeDModelRevisionWrite,
    ThreeDModelUpdate,
    ThreeDModelWrite,
    ThreeDNode,
    ThreeDNodeList,
)
from cognite.client.utils import _json
from cognite.client.utils._auxiliary import interpolate_and_url_encode, split_into_chunks, unpack_items_in_payload
from cognite.client.utils._concurrency import execute_tasks
from cognite.client.utils._identifier import IdentifierSequence, InternalId
from cognite.client.utils._validation import assert_type
from cognite.client.utils.useful_types import SequenceNotStr

if TYPE_CHECKING:
    from cognite.client import CogniteClient
    from cognite.client.config import ClientConfig


class ThreeDAPI(APIClient):
    def __init__(self, config: ClientConfig, api_version: str | None, cognite_client: CogniteClient) -> None:
        super().__init__(config, api_version, cognite_client)
        self.models = ThreeDModelsAPI(config, api_version, cognite_client)
        self.revisions = ThreeDRevisionsAPI(config, api_version, cognite_client)
        self.files = ThreeDFilesAPI(config, api_version, cognite_client)
        self.asset_mappings = ThreeDAssetMappingAPI(config, api_version, cognite_client)


class ThreeDModelsAPI(APIClient):
    _RESOURCE_PATH = "/3d/models"

    @overload
    def __call__(
        self, chunk_size: None = None, published: bool | None = None, limit: int | None = None
    ) -> Iterator[ThreeDModel]: ...

    @overload
    def __call__(
        self, chunk_size: int, published: bool | None = None, limit: int | None = None
    ) -> Iterator[ThreeDModelList]: ...

    def __call__(
        self, chunk_size: int | None = None, published: bool | None = None, limit: int | None = None
    ) -> Iterator[ThreeDModel] | Iterator[ThreeDModelList]:
        """Iterate over 3d models

        Fetches 3d models as they are iterated over, so you keep a limited number of 3d models in memory.

        Args:
            chunk_size (int | None): Number of 3d models to return in each chunk. Defaults to yielding one model a time.
            published (bool | None): Filter based on whether or not the model has published revisions.
            limit (int | None): Maximum number of 3d models to return. Defaults to return all items.

        Returns:
            Iterator[ThreeDModel] | Iterator[ThreeDModelList]: yields ThreeDModel one by one if chunk is not specified, else ThreeDModelList objects.
        """
        return self._list_generator(
            list_cls=ThreeDModelList,
            resource_cls=ThreeDModel,
            method="GET",
            chunk_size=chunk_size,
            filter={"published": published},
            limit=limit,
        )

    def __iter__(self) -> Iterator[ThreeDModel]:
        """Iterate over 3d models

        Fetches models as they are iterated over, so you keep a limited number of models in memory.

        Returns:
            Iterator[ThreeDModel]: yields models one by one.
        """
        return self()

    def retrieve(self, id: int) -> ThreeDModel | None:
        """`Retrieve a 3d model by id <https://developer.cognite.com/api#tag/3D-Models/operation/get3DModel>`_

        Args:
            id (int): Get the model with this id.

        Returns:
            ThreeDModel | None: The requested 3d model.

        Example:

            Get 3d model by id::

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> res = client.three_d.models.retrieve(id=1)
        """
        return self._retrieve(cls=ThreeDModel, identifier=InternalId(id))

    def list(self, published: bool | None = None, limit: int | None = DEFAULT_LIMIT_READ) -> ThreeDModelList:
        """`List 3d models. <https://developer.cognite.com/api#tag/3D-Models/operation/get3DModels>`_

        Args:
            published (bool | None): Filter based on whether or not the model has published revisions.
            limit (int | None): Maximum number of models to retrieve. Defaults to 25. Set to -1, float("inf") or None to return all items.

        Returns:
            ThreeDModelList: The list of 3d models.

        Examples:

            List 3d models:

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> three_d_model_list = client.three_d.models.list()

            Iterate over 3d models:

                >>> for three_d_model in client.three_d.models:
                ...     three_d_model # do something with the 3d model

            Iterate over chunks of 3d models to reduce memory load:

                >>> for three_d_model in client.three_d.models(chunk_size=50):
                ...     three_d_model # do something with the 3d model
        """
        return self._list(
            list_cls=ThreeDModelList,
            resource_cls=ThreeDModel,
            method="GET",
            filter={"published": published},
            limit=limit,
        )

    def create(
        self,
        name: str | ThreeDModelWrite | SequenceNotStr[str | ThreeDModelWrite],
        data_set_id: int | None = None,
        metadata: dict[str, str] | None = None,
    ) -> ThreeDModel | ThreeDModelList:
        """`Create new 3d models. <https://developer.cognite.com/api#tag/3D-Models/operation/create3DModels>`_

        Args:
            name (str | ThreeDModelWrite | SequenceNotStr[str | ThreeDModelWrite]): The name of the 3d model(s) or 3D
                model object to create. If a 3D model object is provided, the other arguments are ignored.
            data_set_id (int | None): The id of the dataset this 3D model belongs to.
            metadata (dict[str, str] | None): Custom, application-specific metadata. String key -> String value.
                Limits: Maximum length of key is 32 bytes, value 512 bytes, up to 16 key-value pairs.

        Returns:
            ThreeDModel | ThreeDModelList: The created 3d model(s).

        Example:

            Create new 3d models::

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> res = client.three_d.models.create(name="My Model", data_set_id=1, metadata={"key1": "value1", "key2": "value2"})

            Create multiple new 3D Models::

                >>> from cognite.client import CogniteClient
                >>> from cognite.client.data_classes import ThreeDModelWrite
                >>> client = CogniteClient()
                >>> my_model = ThreeDModelWrite(name="My Model", data_set_id=1, metadata={"key1": "value1", "key2": "value2"})
                >>> my_other_model = ThreeDModelWrite(name="My Other Model", data_set_id=1, metadata={"key1": "value1", "key2": "value2"})
                >>> res = client.three_d.models.create([my_model, my_other_model])

        """
        items: ThreeDModelWrite | list[ThreeDModelWrite]
        if isinstance(name, str):
            items = ThreeDModelWrite(name, data_set_id, metadata)
        elif isinstance(name, ThreeDModelWrite):
            items = name
        else:
            items = [ThreeDModelWrite(n, data_set_id, metadata) if isinstance(n, str) else n for n in name]
        return self._create_multiple(list_cls=ThreeDModelList, resource_cls=ThreeDModel, items=items)

    @overload
    def update(
        self,
        item: ThreeDModel | ThreeDModelUpdate,
        mode: Literal["replace_ignore_null", "patch", "replace"] = "replace_ignore_null",
    ) -> ThreeDModel: ...

    @overload
    def update(
        self,
        item: Sequence[ThreeDModel | ThreeDModelUpdate],
        mode: Literal["replace_ignore_null", "patch", "replace"] = "replace_ignore_null",
    ) -> ThreeDModelList: ...

    def update(
        self,
        item: ThreeDModel | ThreeDModelUpdate | Sequence[ThreeDModel | ThreeDModelUpdate],
        mode: Literal["replace_ignore_null", "patch", "replace"] = "replace_ignore_null",
    ) -> ThreeDModel | ThreeDModelList:
        """`Update 3d models. <https://developer.cognite.com/api#tag/3D-Models/operation/update3DModels>`_

        Args:
            item (ThreeDModel | ThreeDModelUpdate | Sequence[ThreeDModel | ThreeDModelUpdate]): ThreeDModel(s) to update
            mode (Literal['replace_ignore_null', 'patch', 'replace']): How to update data when a non-update object is given (ThreeDModel or -Write). If you use 'replace_ignore_null', only the fields you have set will be used to replace existing (default). Using 'replace' will additionally clear all the fields that are not specified by you. Last option, 'patch', will update only the fields you have set and for container-like fields such as metadata or labels, add the values to the existing. For more details, see :ref:`appendix-update`.

        Returns:
            ThreeDModel | ThreeDModelList: Updated ThreeDModel(s)

        Examples:

            Update 3d model that you have fetched. This will perform a full update of the model:

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> three_d_model = client.three_d.models.retrieve(id=1)
                >>> three_d_model.name = "New Name"
                >>> res = client.three_d.models.update(three_d_model)

            Perform a partial update on a 3d model:

                >>> from cognite.client.data_classes import ThreeDModelUpdate
                >>> my_update = ThreeDModelUpdate(id=1).name.set("New Name")
                >>> res = client.three_d.models.update(my_update)

        """
        # Note that we cannot use the ThreeDModelWrite to update as the write format of a 3D model
        # does not have ID or External ID, thus no identifier to know which model to update.
        return self._update_multiple(
            list_cls=ThreeDModelList,
            resource_cls=ThreeDModel,
            update_cls=ThreeDModelUpdate,
            items=item,
            mode=mode,
        )

    def delete(self, id: int | Sequence[int]) -> None:
        """`Delete 3d models. <https://developer.cognite.com/api#tag/3D-Models/operation/delete3DModels>`_

        Args:
            id (int | Sequence[int]): ID or list of IDs to delete.

        Example:

            Delete 3d model by id::

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> res = client.three_d.models.delete(id=1)
        """
        self._delete_multiple(identifiers=IdentifierSequence.load(ids=id), wrap_ids=True)


class ThreeDRevisionsAPI(APIClient):
    _RESOURCE_PATH = "/3d/models/{}/revisions"

    @overload
    def __call__(
        self, model_id: int, chunk_size: None = None, published: bool = False, limit: int | None = None
    ) -> Iterator[ThreeDModelRevision]: ...
    @overload
    def __call__(
        self, model_id: int, chunk_size: int, published: bool = False, limit: int | None = None
    ) -> Iterator[ThreeDModelRevisionList]: ...

    def __call__(
        self, model_id: int, chunk_size: int | None = None, published: bool = False, limit: int | None = None
    ) -> Iterator[ThreeDModelRevision] | Iterator[ThreeDModelRevisionList]:
        """Iterate over 3d model revisions

        Fetches 3d model revisions as they are iterated over, so you keep a limited number of 3d model revisions in memory.

        Args:
            model_id (int): Iterate over revisions for the model with this id.
            chunk_size (int | None): Number of 3d model revisions to return in each chunk. Defaults to yielding one model a time.
            published (bool): Filter based on whether or not the revision has been published.
            limit (int | None): Maximum number of 3d model revisions to return. Defaults to return all items.

        Returns:
            Iterator[ThreeDModelRevision] | Iterator[ThreeDModelRevisionList]: yields ThreeDModelRevision one by one if chunk is not specified, else ThreeDModelRevisionList objects.
        """
        return self._list_generator(
            list_cls=ThreeDModelRevisionList,
            resource_cls=ThreeDModelRevision,
            resource_path=interpolate_and_url_encode(self._RESOURCE_PATH, model_id),
            method="GET",
            chunk_size=chunk_size,
            filter={"published": published},
            limit=limit,
        )

    def retrieve(self, model_id: int, id: int) -> ThreeDModelRevision | None:
        """`Retrieve a 3d model revision by id <https://developer.cognite.com/api#tag/3D-Model-Revisions/operation/get3DRevision>`_

        Args:
            model_id (int): Get the revision under the model with this id.
            id (int): Get the model revision with this id.

        Returns:
            ThreeDModelRevision | None: The requested 3d model revision.

        Example:

            Retrieve 3d model revision by model id and revision id::

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> res = client.three_d.revisions.retrieve(model_id=1, id=1)
        """
        return self._retrieve(
            cls=ThreeDModelRevision,
            resource_path=interpolate_and_url_encode(self._RESOURCE_PATH, model_id),
            identifier=InternalId(id),
        )

    @overload
    def create(
        self, model_id: int, revision: ThreeDModelRevision | ThreeDModelRevisionWrite
    ) -> ThreeDModelRevision: ...

    @overload
    def create(
        self, model_id: int, revision: Sequence[ThreeDModelRevision] | Sequence[ThreeDModelRevisionWrite]
    ) -> ThreeDModelRevisionList: ...

    def create(
        self,
        model_id: int,
        revision: ThreeDModelRevision
        | ThreeDModelRevisionWrite
        | Sequence[ThreeDModelRevision]
        | Sequence[ThreeDModelRevisionWrite],
    ) -> ThreeDModelRevision | ThreeDModelRevisionList:
        """`Create a revisions for a specified 3d model. <https://developer.cognite.com/api#tag/3D-Model-Revisions/operation/create3DRevisions>`_

        Args:
            model_id (int): Create revisions for this model.
            revision (ThreeDModelRevision | ThreeDModelRevisionWrite | Sequence[ThreeDModelRevision] | Sequence[ThreeDModelRevisionWrite]): The revision(s) to create.

        Returns:
            ThreeDModelRevision | ThreeDModelRevisionList: The created revision(s)

        Example:

            Create 3d model revision::

                >>> from cognite.client import CogniteClient
                >>> from cognite.client.data_classes import ThreeDModelRevisionWrite
                >>> client = CogniteClient()
                >>> my_revision = ThreeDModelRevisionWrite(file_id=1)
                >>> res = client.three_d.revisions.create(model_id=1, revision=my_revision)
        """
        return self._create_multiple(
            list_cls=ThreeDModelRevisionList,
            resource_cls=ThreeDModelRevision,
            resource_path=interpolate_and_url_encode(self._RESOURCE_PATH, model_id),
            items=revision,
            input_resource_cls=ThreeDModelRevisionWrite,
        )

    def list(
        self, model_id: int, published: bool = False, limit: int | None = DEFAULT_LIMIT_READ
    ) -> ThreeDModelRevisionList:
        """`List 3d model revisions. <https://developer.cognite.com/api#tag/3D-Model-Revisions/operation/get3DRevisions>`_

        Args:
            model_id (int): List revisions under the model with this id.
            published (bool): Filter based on whether or not the revision is published.
            limit (int | None): Maximum number of models to retrieve. Defaults to 25. Set to -1, float("inf") or None to return all items.

        Returns:
            ThreeDModelRevisionList: The list of 3d model revisions.

        Example:

            List 3d model revisions::

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> res = client.three_d.revisions.list(model_id=1, published=True, limit=100)
        """
        return self._list(
            list_cls=ThreeDModelRevisionList,
            resource_cls=ThreeDModelRevision,
            resource_path=interpolate_and_url_encode(self._RESOURCE_PATH, model_id),
            method="GET",
            filter={"published": published},
            limit=limit,
        )

    def update(
        self,
        model_id: int,
        item: ThreeDModelRevision
        | ThreeDModelRevisionUpdate
        | Sequence[ThreeDModelRevision | ThreeDModelRevisionUpdate],
        mode: Literal["replace_ignore_null", "patch", "replace"] = "replace_ignore_null",
    ) -> ThreeDModelRevision | ThreeDModelRevisionList:
        """`Update 3d model revisions. <https://developer.cognite.com/api#tag/3D-Model-Revisions/operation/update3DRevisions>`_

        Args:
            model_id (int): Update the revision under the model with this id.
            item (ThreeDModelRevision | ThreeDModelRevisionUpdate | Sequence[ThreeDModelRevision | ThreeDModelRevisionUpdate]): ThreeDModelRevision(s) to update
            mode (Literal['replace_ignore_null', 'patch', 'replace']): How to update data when a non-update object is given (ThreeDModelRevision or -Write). If you use 'replace_ignore_null', only the fields you have set will be used to replace existing (default). Using 'replace' will additionally clear all the fields that are not specified by you. Last option, 'patch', will update only the fields you have set and for container-like fields such as metadata or labels, add the values to the existing. For more details, see :ref:`appendix-update`.

        Returns:
            ThreeDModelRevision | ThreeDModelRevisionList: Updated ThreeDModelRevision(s)

        Examples:

            Update a revision that you have fetched. This will perform a full update of the revision:

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> revision = client.three_d.revisions.retrieve(model_id=1, id=1)
                >>> revision.status = "New Status"
                >>> res = client.three_d.revisions.update(model_id=1, item=revision)

            Perform a partial update on a revision, updating the published property and adding a new field to metadata:

                >>> from cognite.client.data_classes import ThreeDModelRevisionUpdate
                >>> my_update = ThreeDModelRevisionUpdate(id=1).published.set(False).metadata.add({"key": "value"})
                >>> res = client.three_d.revisions.update(model_id=1, item=my_update)
        """
        return self._update_multiple(
            list_cls=ThreeDModelRevisionList,
            resource_cls=ThreeDModelRevision,
            update_cls=ThreeDModelRevisionUpdate,
            resource_path=interpolate_and_url_encode(self._RESOURCE_PATH, model_id),
            items=item,
            mode=mode,
        )

    def delete(self, model_id: int, id: int | Sequence[int]) -> None:
        """`Delete 3d model revisions. <https://developer.cognite.com/api#tag/3D-Model-Revisions/operation/delete3DRevisions>`_

        Args:
            model_id (int): Delete the revision under the model with this id.
            id (int | Sequence[int]): ID or list of IDs to delete.

        Example:

            Delete 3d model revision by id::

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> res = client.three_d.revisions.delete(model_id=1, id=1)
        """
        self._delete_multiple(
            resource_path=interpolate_and_url_encode(self._RESOURCE_PATH, model_id),
            identifiers=IdentifierSequence.load(ids=id),
            wrap_ids=True,
        )

    def update_thumbnail(self, model_id: int, revision_id: int, file_id: int) -> None:
        """`Update a revision thumbnail. <https://developer.cognite.com/api#tag/3D-Model-Revisions/operation/updateThumbnail>`_

        Args:
            model_id (int): Id of the model.
            revision_id (int): Id of the revision.
            file_id (int): Id of the thumbnail file in the Files API.

        Example:

            Update revision thumbnail::

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> res = client.three_d.revisions.update_thumbnail(model_id=1, revision_id=1, file_id=1)
        """
        resource_path = interpolate_and_url_encode(self._RESOURCE_PATH + "/{}/thumbnail", model_id, revision_id)
        body = {"fileId": file_id}
        self._post(resource_path, json=body)

    def list_nodes(
        self,
        model_id: int,
        revision_id: int,
        node_id: int | None = None,
        depth: int | None = None,
        sort_by_node_id: bool = False,
        partitions: int | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
    ) -> ThreeDNodeList:
        """`Retrieves a list of nodes from the hierarchy in the 3D Model. <https://developer.cognite.com/api#tag/3D-Model-Revisions/operation/get3DNodes>`_

        You can also request a specific subtree with the 'nodeId' query parameter and limit the depth of
        the resulting subtree with the 'depth' query parameter.

        Args:
            model_id (int): Id of the model.
            revision_id (int): Id of the revision.
            node_id (int | None): ID of the root node of the subtree you request (default is the root node).
            depth (int | None): Get sub nodes up to this many levels below the specified node. Depth 0 is the root node.
            sort_by_node_id (bool): Returns the nodes in `nodeId` order.
            partitions (int | None): The result is retrieved in this many parts in parallel. Requires `sort_by_node_id` to be set to `true`.
            limit (int | None): Maximum number of nodes to return. Defaults to 25. Set to -1, float("inf") or None to return all items.

        Returns:
            ThreeDNodeList: The list of 3d nodes.

        Example:

            List nodes from the hierarchy in the 3d model::

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> res = client.three_d.revisions.list_nodes(model_id=1, revision_id=1, limit=10)
        """
        resource_path = interpolate_and_url_encode(self._RESOURCE_PATH + "/{}/nodes", model_id, revision_id)
        return self._list(
            list_cls=ThreeDNodeList,
            resource_cls=ThreeDNode,
            resource_path=resource_path,
            method="GET",
            limit=limit,
            filter={"depth": depth, "nodeId": node_id},
            partitions=partitions,
            other_params={"sortByNodeId": sort_by_node_id},
        )

    def filter_nodes(
        self,
        model_id: int,
        revision_id: int,
        properties: dict[str, dict[str, SequenceNotStr[str]]] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        partitions: int | None = None,
    ) -> ThreeDNodeList:
        """`List nodes in a revision, filtered by node property values. <https://developer.cognite.com/api#tag/3D-Model-Revisions/operation/filter3DNodes>`_

        Args:
            model_id (int): Id of the model.
            revision_id (int): Id of the revision.
            properties (dict[str, dict[str, SequenceNotStr[str]]] | None): Properties for filtering. The object contains one or more category. Each category references one or more properties. Each property is associated with a list of values. For a node to satisfy the filter, it must, for each category/property in the filter, contain the category+property combination with a value that is contained within the corresponding list in the filter.
            limit (int | None): Maximum number of nodes to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            partitions (int | None): The result is retrieved in this many parts in parallel. Requires `sort_by_node_id` to be set to `true`.

        Returns:
            ThreeDNodeList: The list of 3d nodes.

        Example:

            Filter nodes from the hierarchy in the 3d model that have one of the values "AB76", "AB77" or "AB78" for property PDMS/Area AND that also have one of the values "PIPE", "BEND" or "PIPESUP" for the property PDMS/Type.

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> res = client.three_d.revisions.filter_nodes(model_id=1, revision_id=1, properties={ "PDMS": { "Area": ["AB76", "AB77", "AB78"], "Type": ["PIPE", "BEND", "PIPESUP"] } }, limit=10)
        """
        resource_path = interpolate_and_url_encode(self._RESOURCE_PATH + "/{}/nodes", model_id, revision_id)
        return self._list(
            list_cls=ThreeDNodeList,
            resource_cls=ThreeDNode,
            resource_path=resource_path,
            method="POST",
            limit=limit,
            filter={"properties": properties},
            partitions=partitions,
        )

    def list_ancestor_nodes(
        self, model_id: int, revision_id: int, node_id: int | None = None, limit: int | None = DEFAULT_LIMIT_READ
    ) -> ThreeDNodeList:
        """`Retrieves a list of ancestor nodes of a given node, including itself, in the hierarchy of the 3D model <https://developer.cognite.com/api#tag/3D-Model-Revisions/operation/get3DNodeAncestors>`_

        Args:
            model_id (int): Id of the model.
            revision_id (int): Id of the revision.
            node_id (int | None): ID of the node to get the ancestors of.
            limit (int | None): Maximum number of nodes to return. Defaults to 25. Set to -1, float("inf") or None to return all items.

        Returns:
            ThreeDNodeList: The list of 3d nodes.

        Example:

            Get a list of ancestor nodes of a given node::

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> res = client.three_d.revisions.list_ancestor_nodes(model_id=1, revision_id=1, node_id=5, limit=10)
        """
        resource_path = interpolate_and_url_encode(self._RESOURCE_PATH + "/{}/nodes", model_id, revision_id)
        return self._list(
            list_cls=ThreeDNodeList,
            resource_cls=ThreeDNode,
            resource_path=resource_path,
            method="GET",
            limit=limit,
            filter={"nodeId": node_id},
        )


class ThreeDFilesAPI(APIClient):
    _RESOURCE_PATH = "/3d/files"

    def retrieve(self, id: int) -> bytes:
        """`Retrieve the contents of a 3d file by id. <https://developer.cognite.com/api#tag/3D-Files/operation/get3DFile>`_

        Args:
            id (int): The id of the file to retrieve.

        Returns:
            bytes: The contents of the file.

        Example:

            Retrieve the contents of a 3d file by id::

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> res = client.three_d.files.retrieve(1)
        """
        path = interpolate_and_url_encode(self._RESOURCE_PATH + "/{}", id)
        return self._get(path).content


class ThreeDAssetMappingAPI(APIClient):
    _RESOURCE_PATH = "/3d/models/{}/revisions/{}/mappings"

    def list(
        self,
        model_id: int,
        revision_id: int,
        node_id: int | None = None,
        asset_id: int | None = None,
        intersects_bounding_box: BoundingBox3D | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
    ) -> ThreeDAssetMappingList:
        """`List 3D node asset mappings. <https://developer.cognite.com/api#tag/3D-Asset-Mapping/operation/get3DMappings>`_

        Args:
            model_id (int): Id of the model.
            revision_id (int): Id of the revision.
            node_id (int | None): List only asset mappings associated with this node.
            asset_id (int | None): List only asset mappings associated with this asset.
            intersects_bounding_box (BoundingBox3D | None): If given, only return asset mappings for assets whose bounding box intersects with the given bounding box.
            limit (int | None): Maximum number of asset mappings to return. Defaults to 25. Set to -1, float("inf") or None to return all items.

        Returns:
            ThreeDAssetMappingList: The list of asset mappings.

        Example:

            List 3d node asset mappings:

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> res = client.three_d.asset_mappings.list(model_id=1, revision_id=1)

            List 3d node asset mappings for assets whose bounding box intersects with a given bounding box:

                >>> from cognite.client.data_classes import BoundingBox3D
                >>> bbox = BoundingBox3D(min=[0.0, 0.0, 0.0], max=[1.0, 1.0, 1.0])
                >>> res = client.three_d.asset_mappings.list(
                ...     model_id=1, revision_id=1, intersects_bounding_box=bbox)
        """
        path = interpolate_and_url_encode(self._RESOURCE_PATH, model_id, revision_id)
        flt: dict[str, str | int | None] = {"nodeId": node_id, "assetId": asset_id}
        if intersects_bounding_box:
            flt["intersectsBoundingBox"] = _json.dumps(intersects_bounding_box)
        return self._list(
            list_cls=ThreeDAssetMappingList,
            resource_cls=ThreeDAssetMapping,
            resource_path=path,
            method="GET",
            filter=flt,
            limit=limit,
        )

    @overload
    def create(
        self, model_id: int, revision_id: int, asset_mapping: ThreeDAssetMapping | ThreeDAssetMappingWrite
    ) -> ThreeDAssetMapping: ...

    @overload
    def create(
        self,
        model_id: int,
        revision_id: int,
        asset_mapping: Sequence[ThreeDAssetMapping] | Sequence[ThreeDAssetMappingWrite],
    ) -> ThreeDAssetMappingList: ...

    def create(
        self,
        model_id: int,
        revision_id: int,
        asset_mapping: ThreeDAssetMapping
        | ThreeDAssetMappingWrite
        | Sequence[ThreeDAssetMapping]
        | Sequence[ThreeDAssetMappingWrite],
    ) -> ThreeDAssetMapping | ThreeDAssetMappingList:
        """`Create 3d node asset mappings. <https://developer.cognite.com/api#tag/3D-Asset-Mapping/operation/create3DMappings>`_

        Args:
            model_id (int): Id of the model.
            revision_id (int): Id of the revision.
            asset_mapping (ThreeDAssetMapping | ThreeDAssetMappingWrite | Sequence[ThreeDAssetMapping] | Sequence[ThreeDAssetMappingWrite]): The asset mapping(s) to create.

        Returns:
            ThreeDAssetMapping | ThreeDAssetMappingList: The created asset mapping(s).

        Example:

            Create new 3d node asset mapping::

                >>> from cognite.client import CogniteClient
                >>> from cognite.client.data_classes import ThreeDAssetMappingWrite
                >>> my_mapping = ThreeDAssetMappingWrite(node_id=1, asset_id=1)
                >>> client = CogniteClient()
                >>> res = client.three_d.asset_mappings.create(model_id=1, revision_id=1, asset_mapping=my_mapping)
        """
        path = interpolate_and_url_encode(self._RESOURCE_PATH, model_id, revision_id)
        return self._create_multiple(
            list_cls=ThreeDAssetMappingList,
            resource_cls=ThreeDAssetMapping,
            resource_path=path,
            items=asset_mapping,
            input_resource_cls=ThreeDAssetMappingWrite,
        )

    def delete(
        self, model_id: int, revision_id: int, asset_mapping: ThreeDAssetMapping | Sequence[ThreeDAssetMapping]
    ) -> None:
        """`Delete 3d node asset mappings. <https://developer.cognite.com/api#tag/3D-Asset-Mapping/operation/delete3DMappings>`_

        Args:
            model_id (int): Id of the model.
            revision_id (int): Id of the revision.
            asset_mapping (ThreeDAssetMapping | Sequence[ThreeDAssetMapping]): The asset mapping(s) to delete.

        Example:

            Delete 3d node asset mapping::

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> mapping_to_delete = client.three_d.asset_mappings.list(model_id=1, revision_id=1)[0]
                >>> res = client.three_d.asset_mappings.delete(model_id=1, revision_id=1, asset_mapping=mapping_to_delete)
        """
        path = interpolate_and_url_encode(self._RESOURCE_PATH, model_id, revision_id)
        assert_type(asset_mapping, "asset_mapping", [Sequence, ThreeDAssetMapping])
        if isinstance(asset_mapping, ThreeDAssetMapping):
            asset_mapping = [asset_mapping]
        chunks = split_into_chunks(
            [ThreeDAssetMapping(a.node_id, a.asset_id).dump(camel_case=True) for a in asset_mapping], self._DELETE_LIMIT
        )
        tasks = [{"url_path": path + "/delete", "json": {"items": chunk}} for chunk in chunks]
        summary = execute_tasks(self._post, tasks, self._config.max_workers)
        summary.raise_compound_exception_if_failed_tasks(
            task_unwrap_fn=unpack_items_in_payload, task_list_element_unwrap_fn=lambda el: ThreeDAssetMapping._load(el)
        )



from __future__ import annotations

from collections.abc import Iterator, Sequence
from typing import TYPE_CHECKING, overload

from cognite.client._api_client import APIClient
from cognite.client._constants import DEFAULT_LIMIT_READ
from cognite.client.data_classes.simulators.filters import SimulatorRunsFilter
from cognite.client.data_classes.simulators.runs import (
    SimulationRun,
    SimulationRunWrite,
    SimulatorRunDataList,
    SimulatorRunList,
)
from cognite.client.utils._experimental import FeaturePreviewWarning
from cognite.client.utils._identifier import IdentifierSequence
from cognite.client.utils._validation import assert_type
from cognite.client.utils.useful_types import SequenceNotStr

if TYPE_CHECKING:
    from cognite.client import CogniteClient
    from cognite.client.config import ClientConfig


class SimulatorRunsAPI(APIClient):
    _RESOURCE_PATH = "/simulators/runs"
    _RESOURCE_PATH_RUN = "/simulators/run"

    def __init__(
        self,
        config: ClientConfig,
        api_version: str | None,
        cognite_client: CogniteClient,
    ) -> None:
        super().__init__(config, api_version, cognite_client)
        self._CREATE_LIMIT = 1
        self._RETRIEVE_LIMIT = 1
        self._warning = FeaturePreviewWarning(
            api_maturity="General Availability",
            sdk_maturity="alpha",
            feature_name="Simulators",
        )

    def __iter__(self) -> Iterator[SimulationRun]:
        """Iterate over simulation runs

        Fetches simulation runs as they are iterated over, so you keep a limited number of simulation runs in memory.

        Returns:
            Iterator[SimulationRun]: yields simulation runs one by one.
        """
        return self()

    @overload
    def __call__(
        self,
        chunk_size: int,
        limit: int | None = None,
        status: str | None = None,
        run_type: str | None = None,
        model_external_ids: SequenceNotStr[str] | None = None,
        simulator_integration_external_ids: SequenceNotStr[str] | None = None,
        simulator_external_ids: SequenceNotStr[str] | None = None,
        routine_external_ids: SequenceNotStr[str] | None = None,
        routine_revision_external_ids: SequenceNotStr[str] | None = None,
        model_revision_external_ids: SequenceNotStr[str] | None = None,
    ) -> Iterator[SimulatorRunList]: ...

    @overload
    def __call__(
        self,
        chunk_size: None = None,
        limit: int | None = None,
        status: str | None = None,
        run_type: str | None = None,
        model_external_ids: SequenceNotStr[str] | None = None,
        simulator_integration_external_ids: SequenceNotStr[str] | None = None,
        simulator_external_ids: SequenceNotStr[str] | None = None,
        routine_external_ids: SequenceNotStr[str] | None = None,
        routine_revision_external_ids: SequenceNotStr[str] | None = None,
        model_revision_external_ids: SequenceNotStr[str] | None = None,
    ) -> Iterator[SimulationRun]: ...

    def __call__(
        self,
        chunk_size: int | None = None,
        limit: int | None = None,
        status: str | None = None,
        run_type: str | None = None,
        model_external_ids: SequenceNotStr[str] | None = None,
        simulator_integration_external_ids: SequenceNotStr[str] | None = None,
        simulator_external_ids: SequenceNotStr[str] | None = None,
        routine_external_ids: SequenceNotStr[str] | None = None,
        routine_revision_external_ids: SequenceNotStr[str] | None = None,
        model_revision_external_ids: SequenceNotStr[str] | None = None,
    ) -> Iterator[SimulationRun] | Iterator[SimulatorRunList]:
        """Iterate over simulation runs

        Fetches simulation runs as they are iterated over, so you keep a limited number of simulation runs in memory.

        Args:
            chunk_size (int | None): Number of simulation runs to return in each chunk. Defaults to yielding one simulation run a time.
            limit (int | None): The maximum number of simulation runs to return, pass None to return all.
            status (str | None): Filter by simulation run status
            run_type (str | None): Filter by simulation run type
            model_external_ids (SequenceNotStr[str] | None): Filter by simulator model external ids
            simulator_integration_external_ids (SequenceNotStr[str] | None): Filter by simulator integration external ids
            simulator_external_ids (SequenceNotStr[str] | None): Filter by simulator external ids
            routine_external_ids (SequenceNotStr[str] | None): Filter by routine external ids
            routine_revision_external_ids (SequenceNotStr[str] | None): Filter by routine revision external ids
            model_revision_external_ids (SequenceNotStr[str] | None): Filter by model revision external ids

        Returns:
            Iterator[SimulationRun] | Iterator[SimulatorRunList]: yields Simulation Run one by one if chunk is not specified, else SimulatorRunsList objects.
        """

        filter_runs = SimulatorRunsFilter(
            status=status,
            run_type=run_type,
            model_external_ids=model_external_ids,
            simulator_integration_external_ids=simulator_integration_external_ids,
            simulator_external_ids=simulator_external_ids,
            routine_external_ids=routine_external_ids,
            routine_revision_external_ids=routine_revision_external_ids,
            model_revision_external_ids=model_revision_external_ids,
        )

        return self._list_generator(
            list_cls=SimulatorRunList,
            resource_cls=SimulationRun,
            method="POST",
            filter=filter_runs.dump(),
            chunk_size=chunk_size,
            limit=limit,
        )

    def list(
        self,
        limit: int | None = DEFAULT_LIMIT_READ,
        status: str | None = None,
        run_type: str | None = None,
        model_external_ids: SequenceNotStr[str] | None = None,
        simulator_integration_external_ids: SequenceNotStr[str] | None = None,
        simulator_external_ids: SequenceNotStr[str] | None = None,
        routine_external_ids: SequenceNotStr[str] | None = None,
        routine_revision_external_ids: SequenceNotStr[str] | None = None,
        model_revision_external_ids: SequenceNotStr[str] | None = None,
    ) -> SimulatorRunList:
        """`Filter simulation runs <https://developer.cognite.com/api#tag/Simulation-Runs/operation/filter_simulation_runs_simulators_runs_list_post>`_

        Retrieves a list of simulation runs that match the given criteria.

        Args:
            limit (int | None): The maximum number of simulation runs to return, pass None to return all.
            status (str | None): Filter by simulation run status
            run_type (str | None): Filter by simulation run type
            model_external_ids (SequenceNotStr[str] | None): Filter by simulator model external ids
            simulator_integration_external_ids (SequenceNotStr[str] | None): Filter by simulator integration external ids
            simulator_external_ids (SequenceNotStr[str] | None): Filter by simulator external ids
            routine_external_ids (SequenceNotStr[str] | None): Filter by routine external ids
            routine_revision_external_ids (SequenceNotStr[str] | None): Filter by routine revision external ids
            model_revision_external_ids (SequenceNotStr[str] | None): Filter by model revision external ids

        Returns:
            SimulatorRunList: List of simulation runs

        Examples:
            List simulation runs:
                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> res = client.simulators.runs.list()

            Filter runs by status and simulator external ids:
                >>> res = client.simulators.runs.list(
                ...     simulator_external_ids=["PROSPER", "DWSIM"],
                ...     status="success"
                ... )
        """

        filter_runs = SimulatorRunsFilter(
            status=status,
            run_type=run_type,
            model_external_ids=model_external_ids,
            simulator_integration_external_ids=simulator_integration_external_ids,
            simulator_external_ids=simulator_external_ids,
            routine_external_ids=routine_external_ids,
            routine_revision_external_ids=routine_revision_external_ids,
            model_revision_external_ids=model_revision_external_ids,
        )
        self._warning.warn()
        return self._list(
            method="POST",
            limit=limit,
            resource_cls=SimulationRun,
            list_cls=SimulatorRunList,
            filter=filter_runs.dump(),
        )

    @overload
    def retrieve(self, ids: int) -> SimulationRun | None: ...

    @overload
    def retrieve(
        self,
        ids: Sequence[int],
    ) -> SimulatorRunList | None: ...

    def retrieve(
        self,
        ids: int | Sequence[int],
    ) -> SimulationRun | SimulatorRunList | None:
        """`Retrieve simulation runs by ID <https://api-docs.cognite.com/20230101/tag/Simulation-Runs/operation/simulation_by_id_simulators_runs_byids_post>`_

        Args:
            ids (int | Sequence[int]): The ID(s) of the simulation run(s) to retrieve.

        Returns:
            SimulationRun | SimulatorRunList | None: The simulation run(s) with the given ID(s)

        Examples:
            Retrieve a single simulation run by id:
                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> run = client.simulators.runs.retrieve(ids=2)
        """
        self._warning.warn()
        identifiers = IdentifierSequence.load(ids=ids)
        return self._retrieve_multiple(
            resource_cls=SimulationRun,
            list_cls=SimulatorRunList,
            identifiers=identifiers,
            resource_path=self._RESOURCE_PATH,
        )

    @overload
    def create(self, items: SimulationRunWrite) -> SimulationRun: ...

    @overload
    def create(self, items: Sequence[SimulationRunWrite]) -> SimulatorRunList: ...

    def create(self, items: SimulationRunWrite | Sequence[SimulationRunWrite]) -> SimulationRun | SimulatorRunList:
        """`Create simulation runs <https://developer.cognite.com/api#tag/Simulation-Runs/operation/run_simulation_simulators_run_post>`_

        Args:
            items (SimulationRunWrite | Sequence[SimulationRunWrite]): The simulation run(s) to execute.

        Returns:
            SimulationRun | SimulatorRunList: Created simulation run(s)

        Examples:
            Create new simulation run:
                >>> from cognite.client import CogniteClient
                >>> from cognite.client.data_classes.simulators.runs import SimulationRunWrite
                >>> client = CogniteClient()
                >>> run = [
                ...     SimulationRunWrite(
                ...         routine_external_id="routine1",
                ...         log_severity="Debug",
                ...         run_type="external",
                ...     ),
                ... ]
                >>> res = client.simulators.runs.create(run)
        """
        assert_type(items, "simulation_run", [SimulationRunWrite, Sequence])

        return self._create_multiple(
            list_cls=SimulatorRunList,
            resource_cls=SimulationRun,
            items=items,
            input_resource_cls=SimulationRunWrite,
            resource_path=self._RESOURCE_PATH_RUN,
        )

    def list_run_data(
        self,
        run_id: int,
    ) -> SimulatorRunDataList:
        """`Get simulation run data <https://developer.cognite.com/api#tag/Simulation-Runs/operation/simulation_data_by_run_id_simulators_runs_data_list_post>`_

        Retrieve data associated with a simulation run by ID.

        Args:
            run_id (int): Simulation run id.

        Returns:
            SimulatorRunDataList: List of simulation run data

        Examples:
            Get simulation run data by run id:
                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> res = client.simulators.runs.list_run_data(run_id=12345)

            Get simulation run data directly on a simulation run object:
                >>> run = client.simulators.runs.retrieve(ids=2)
                >>> res = run.get_data()
        """
        self._warning.warn()

        req = self._post(
            url_path=f"{self._RESOURCE_PATH}/data/list",
            json={"items": [{"runId": run_id}]},
        )

        items = req.json().get("items", [])
        return SimulatorRunDataList._load(items, cognite_client=self._cognite_client)



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, ClassVar, Literal

if TYPE_CHECKING:
    from cognite.client import CogniteClient

from cognite.client.data_classes._base import (
    CogniteResourceList,
    WriteableCogniteResource,
    WriteableCogniteResourceList,
)


@dataclass
class AgentToolCore(WriteableCogniteResource["AgentToolUpsert"], ABC):
    """Core representation of an AI Agent Tool in CDF.

    Args:
        name (str): The name of the agent tool. Used by the agent to decide when to use this tool.
        description (str): The description of the agent tool. Used by the agent to decide when to use this tool.
    """

    _type: ClassVar[str]  # Will be set by concrete classes
    name: str
    description: str

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        """Dump the instance into a json serializable Python data type."""
        result = super().dump(camel_case=camel_case)

        if hasattr(self.__class__, "_type"):
            result["type"] = self._type
        return result


@dataclass
class AgentToolUpsert(AgentToolCore, ABC):
    """The write format of an agent tool.

    Args:
        name (str): The name of the agent tool. Used by the agent to decide when to use this tool.
        description (str): The description of the agent tool. Used by the agent to decide when to use this tool.
    """

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        """Dump the instance into a json serializable Python data type."""
        result = super().dump(camel_case=camel_case)
        # Use the _type from the class if available, otherwise use instance type
        result["type"] = getattr(self.__class__, "_type", getattr(self, "type", ""))
        return result

    def as_write(self) -> AgentToolUpsert:
        return self

    @classmethod
    def _load_tool(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> AgentToolUpsert:
        return cls(
            name=resource["name"],
            description=resource["description"],
        )


@dataclass
class AgentTool(AgentToolCore, ABC):
    """The read format of an agent tool.

    Args:
        name (str): The name of the agent tool. Used by the agent to decide when to use this tool.
        description (str): The description of the agent tool. Used by the agent to decide when to use this tool.
    """

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        """Dump the instance into a json serializable Python data type."""
        result = super().dump(camel_case=camel_case)
        # Use the _type from the class if available, otherwise use instance type
        result["type"] = getattr(self.__class__, "_type", getattr(self, "type", ""))
        return result

    @abstractmethod
    def as_write(self) -> AgentToolUpsert:
        """Convert this tool to its upsert representation."""
        pass

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> AgentTool:
        tool_type = resource.get("type", "")

        # Get the appropriate class based on the tool.type
        tool_class = _AGENT_TOOL_CLS_BY_TYPE.get(tool_type, UnknownAgentTool)

        return tool_class._load_tool(resource, cognite_client)

    @classmethod
    def _load_tool(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> AgentTool:
        """Default instance loading method for simple tool types."""
        return cls(
            name=resource["name"],
            description=resource["description"],
        )


@dataclass
class DataModelInfo(WriteableCogniteResource):
    """Information about a data model used in knowledge graph queries.

    Args:
        space (str): The space of the data model.
        external_id (str): The external ID of the data model.
        version (str): The version of the data model.
        view_external_ids (Sequence[str] | None): The external IDs of the views of the data model.
    """

    space: str
    external_id: str
    version: str
    view_external_ids: Sequence[str] | None = None

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> DataModelInfo:
        return cls(
            space=resource["space"],
            external_id=resource["externalId"],
            version=resource["version"],
            view_external_ids=resource.get("viewExternalIds"),
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        result = super().dump(camel_case=camel_case)
        return result

    def as_write(self) -> DataModelInfo:
        return self


@dataclass
class InstanceSpaces(WriteableCogniteResource):
    """Configuration for instance spaces in knowledge graph queries.

    Args:
        type (Literal["manual", "all"]): The type of instance spaces.
        spaces (Sequence[str] | None): The spaces of the instance spaces.
    """

    type: Literal["manual", "all"]
    spaces: Sequence[str] | None = None

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> InstanceSpaces:
        return cls(
            type=resource["type"],
            spaces=resource.get("spaces"),
        )

    def as_write(self) -> InstanceSpaces:
        return self


@dataclass
class QueryKnowledgeGraphAgentToolConfiguration(WriteableCogniteResource):
    """Configuration for knowledge graph query agent tools.

    Args:
        data_models (Sequence[DataModelInfo] | None): The data models and views to query.
        instance_spaces (InstanceSpaces | None): The instance spaces to query.
    """

    data_models: Sequence[DataModelInfo] | None = None
    instance_spaces: InstanceSpaces | None = None

    @classmethod
    def _load(
        cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None
    ) -> QueryKnowledgeGraphAgentToolConfiguration:
        data_models = None
        if "dataModels" in resource:
            data_models = [DataModelInfo._load(dm) for dm in resource["dataModels"]]

        instance_spaces = None
        if "instanceSpaces" in resource:
            instance_spaces = InstanceSpaces._load(resource["instanceSpaces"])

        return cls(
            data_models=data_models,
            instance_spaces=instance_spaces,
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        result: dict[str, Any] = {}
        if self.data_models:
            key = "dataModels" if camel_case else "data_models"
            result[key] = [dm.dump(camel_case=camel_case) for dm in self.data_models]
        if self.instance_spaces:
            key = "instanceSpaces" if camel_case else "instance_spaces"
            result[key] = self.instance_spaces.dump(camel_case=camel_case)
        return result

    def as_write(self) -> QueryKnowledgeGraphAgentToolConfiguration:
        return self


@dataclass
class SummarizeDocumentAgentTool(AgentTool):
    """Agent tool for summarizing documents.

    Args:
        name (str): The name of the agent tool. Used by the agent to decide when to use this tool.
        description (str): The description of the agent tool. Used by the agent to decide when to use this tool.
    """

    _type: ClassVar[str] = "summarizeDocument"

    def as_write(self) -> SummarizeDocumentAgentToolUpsert:
        return SummarizeDocumentAgentToolUpsert(
            name=self.name,
            description=self.description,
        )

    @classmethod
    def _load_tool(
        cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None
    ) -> SummarizeDocumentAgentTool:
        return cls(
            name=resource["name"],
            description=resource["description"],
        )


@dataclass
class SummarizeDocumentAgentToolUpsert(AgentToolUpsert):
    """Upsert version of document summarization agent tool.

    Args:
        name (str): The name of the agent tool. Used by the agent to decide when to use this tool.
        description (str): The description of the agent tool. Used by the agent to decide when to use this tool.
    """

    _type: ClassVar[str] = "summarizeDocument"

    @classmethod
    def _load(
        cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None
    ) -> SummarizeDocumentAgentToolUpsert:
        return cls(
            name=resource["name"],
            description=resource["description"],
        )


@dataclass
class AskDocumentAgentTool(AgentTool):
    """Agent tool for asking questions about documents.

    Args:
        name (str): The name of the agent tool. Used by the agent to decide when to use this tool.
        description (str): The description of the agent tool. Used by the agent to decide when to use this tool.
    """

    _type: ClassVar[str] = "askDocument"

    def as_write(self) -> AskDocumentAgentToolUpsert:
        return AskDocumentAgentToolUpsert(
            name=self.name,
            description=self.description,
        )

    @classmethod
    def _load_tool(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> AskDocumentAgentTool:
        return cls(
            name=resource["name"],
            description=resource["description"],
        )


@dataclass
class AskDocumentAgentToolUpsert(AgentToolUpsert):
    """Upsert version of document question agent tool.

    Args:
        name (str): The name of the agent tool. Used by the agent to decide when to use this tool.
        description (str): The description of the agent tool. Used by the agent to decide when to use this tool.
    """

    _type: ClassVar[str] = "askDocument"

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> AskDocumentAgentToolUpsert:
        return cls(
            name=resource["name"],
            description=resource["description"],
        )


@dataclass
class QueryKnowledgeGraphAgentTool(AgentTool):
    """Agent tool for querying knowledge graphs.

    Args:
        name (str): The name of the agent tool. Used by the agent to decide when to use this tool.
        description (str): The description of the agent tool. Used by the agent to decide when to use this tool.
        configuration (QueryKnowledgeGraphAgentToolConfiguration | None): The configuration of the knowledge graph query agent tool.
    """

    _type: ClassVar[str] = "queryKnowledgeGraph"
    configuration: QueryKnowledgeGraphAgentToolConfiguration | None = None

    @classmethod
    def _load_tool(
        cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None
    ) -> QueryKnowledgeGraphAgentTool:
        # Parse the configuration specifically for this tool type
        configuration = None
        if resource.get("configuration"):
            configuration = QueryKnowledgeGraphAgentToolConfiguration._load(resource["configuration"])

        return cls(
            name=resource["name"],
            description=resource["description"],
            configuration=configuration,
        )

    def as_write(self) -> QueryKnowledgeGraphAgentToolUpsert:
        return QueryKnowledgeGraphAgentToolUpsert(
            name=self.name,
            description=self.description,
            configuration=self.configuration,
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        result = super().dump(camel_case=camel_case)
        if self.configuration:
            result["configuration"] = self.configuration.dump(camel_case=camel_case)
        return result


@dataclass
class QueryKnowledgeGraphAgentToolUpsert(AgentToolUpsert):
    """Upsert version of knowledge graph query agent tool.

    Args:
        name (str): The name of the agent tool. Used by the agent to decide when to use this tool.
        description (str): The description of the agent tool. Used by the agent to decide when to use this tool.
        configuration (QueryKnowledgeGraphAgentToolConfiguration | None): The configuration of the knowledge graph query agent tool.
    """

    _type: ClassVar[str] = "queryKnowledgeGraph"
    configuration: QueryKnowledgeGraphAgentToolConfiguration | None = None

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        result = super().dump(camel_case=camel_case)
        if self.configuration:
            result["configuration"] = self.configuration.dump(camel_case=camel_case)
        return result

    @classmethod
    def _load(
        cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None
    ) -> QueryKnowledgeGraphAgentToolUpsert:
        configuration = None
        if resource.get("configuration"):
            configuration = QueryKnowledgeGraphAgentToolConfiguration._load(resource["configuration"])

        return cls(
            name=resource["name"],
            description=resource["description"],
            configuration=configuration,
        )


@dataclass
class QueryTimeSeriesDatapointsAgentTool(AgentTool):
    """Agent tool for querying time series datapoints.

    Args:
        name (str): The name of the agent tool. Used by the agent to decide when to use this tool.
        description (str): The description of the agent tool. Used by the agent to decide when to use this tool.
    """

    _type: ClassVar[str] = "queryTimeSeriesDatapoints"

    def as_write(self) -> QueryTimeSeriesDatapointsAgentToolUpsert:
        return QueryTimeSeriesDatapointsAgentToolUpsert(
            name=self.name,
            description=self.description,
        )

    @classmethod
    def _load_tool(
        cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None
    ) -> QueryTimeSeriesDatapointsAgentTool:
        return cls(
            name=resource["name"],
            description=resource["description"],
        )


@dataclass
class QueryTimeSeriesDatapointsAgentToolUpsert(AgentToolUpsert):
    """Upsert version of time series datapoints query agent tool.

    Args:
        name (str): The name of the agent tool. Used by the agent to decide when to use this tool.
        description (str): The description of the agent tool. Used by the agent to decide when to use this tool.
    """

    _type: ClassVar[str] = "queryTimeSeriesDatapoints"

    @classmethod
    def _load(
        cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None
    ) -> QueryTimeSeriesDatapointsAgentToolUpsert:
        return cls(
            name=resource["name"],
            description=resource["description"],
        )


@dataclass
class UnknownAgentTool(AgentTool):
    """Agent tool for unknown/unrecognized tool types.

    Args:
        name (str): The name of the agent tool. Used by the agent to decide when to use this tool.
        description (str): The description of the agent tool. Used by the agent to decide when to use this tool.
        type (str): The type of the agent tool.
        configuration (dict[str, Any] | None): The configuration of the agent tool.
    """

    type: str
    configuration: dict[str, Any] | None = None

    def as_write(self) -> UnknownAgentToolUpsert:
        return UnknownAgentToolUpsert(
            name=self.name,
            type=self.type,
            description=self.description,
            configuration=self.configuration,
        )

    @classmethod
    def _load_tool(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> UnknownAgentTool:
        return cls(
            name=resource["name"],
            type=resource["type"],
            description=resource["description"],
            configuration=resource.get("configuration"),
        )


@dataclass
class UnknownAgentToolUpsert(AgentToolUpsert):
    """Upsert version of unknown agent tool.

    Args:
        name (str): The name of the agent tool. Used by the agent to decide when to use this tool.
        type (str): The type of the agent tool.
        description (str): The description of the agent tool. Used by the agent to decide when to use this tool.
        configuration (dict[str, Any] | None): The configuration of the agent tool.
    """

    type: str
    configuration: dict[str, Any] | None = None

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> UnknownAgentToolUpsert:
        return cls(
            name=resource["name"],
            type=resource["type"],
            description=resource["description"],
            configuration=resource.get("configuration"),
        )


class AgentToolUpsertList(CogniteResourceList[AgentToolUpsert]):
    _RESOURCE = AgentToolUpsert


class AgentToolList(WriteableCogniteResourceList[AgentToolUpsert, AgentTool]):
    _RESOURCE = AgentTool

    def as_write(self) -> AgentToolUpsertList:
        """Returns this agent tool list as a writeable instance"""
        return AgentToolUpsertList([item.as_write() for item in self.data], cognite_client=self._get_cognite_client())


# Build the mapping AFTER all classes are defined
_AGENT_TOOL_CLS_BY_TYPE: dict[str, type[AgentTool]] = {
    subclass._type: subclass  # type: ignore[type-abstract]
    for subclass in AgentTool.__subclasses__()
    if hasattr(subclass, "_type") and not getattr(subclass, "__abstractmethods__", None)
}
